;;;-*- Mode: Lisp; Package: DATABROWSER -*-

;;   Copyright (C) 2003 Sandia National Laboratories

;; Sandia National Laboratories grants you the rights to distribute
;; and use this software as governed by the terms
;; of the Lisp Lesser GNU Public License
;; (http://opensource.franz.com/preamble.html),
;; known as the LLGPL.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; Lesser General Public License for more details.

;;; Databrowser.Lisp
;;; Version 1.0

; THIS FILE MUST BE LOADED, NOT MERELY EXECUTED.

;;; Developed by Shannon Spires (svspire@sandia.gov) with the gracious assistance of Terje Norderhaug (xpro@mac.com). 

;;; Check for updates at http://aisl.sandia.gov/opensource/

;;; This will only work in carbonized MCL (4.4b4 and later).
;;; And it looks a lot better in OSX native than in Classic or Carbon on OS9.
;;; See ReadMe in BasicDataBrowser sample code in CarbonLib_1.4 SDK or later.
;;; Also see <a href="http://developer.apple.com/technotes/tn/tn2009.html">Technote 2009</a>
;;; NB: this Technote contains several errors.

;;; See Apple Technote 2009 for more info, but beware: the December 2000 version contains
;;;   several typos. Huge swaths of the databrowser API are also not documented in this
;;;   technote, nor anywhere else unless you count being mentioned by name in the Control
;;;   Manager documentation. Big help.
;;;  http://developer.apple.com/technotes/tn/tn2009.html

;;; History

;;; 5/23/2002 TN and SVS got basic functionality under MCL 4.4b3 on OSX.
;;; 5/31/2002 TN got it almost working on OS9, but column titles still look weird.
;;; 6/2/2002  SVS cleaned up, made collection-databrowser (in another file) which has a higher-level API.
;;;           3 property-types are supported: :TEXT, :TIME, and :CHECKBOX. Only list-view (not column-view)
;;;           is supported for now. Let's walk before we run.
;;; 6/3/2002  SVS attempting to get comparisons working. Clicking on a column header has no effect; the only
;;;           column that seems to be sortable is the farthest-left, regardless of which one you click on.
;;;           Our comparison callbacks are definitely getting called--they're just not making any difference.
;;;           Calling #_getdatabrowsersortproperty confirms that the clicked-on column is the one
;;;           being sorted on, and tracing lookup-browser-data confirms that the right data is
;;;           being looked at by the comparer. It's just not having any effect. Only thing I can
;;;           think of is that defpascals that return booleans need to have some heckuva special
;;;           return value. I've tried T and NIL and also #$true and #$false.
;;; 6/8/02    TN got databrowser-compare-proc working.
;;; 6/9/02    SVS added error handling to defpascals so Lisp functions called therein are less likely
;;;           to crash MCL. :TIME columns a little more robust. databrowser-compare is now more robust
;;;           if one item is nil, etc.
;;; 7/2/02    SVS Editing of fields is now possible. The trick was to add the method on
;;;           %databrowser-item-data ((browser databrowser) rowID (property (eql #$kDataBrowserItemIsEditableProperty)) item-data set?)
;;;           This trick only seems to work when the code is run native on OSX. It doesn't work when you
;;;           boot from OS9. With Carbonlib 1.4, it has no effect--you still can't get an editing frame around
;;;           the cell you're trying to edit. With Carbonlib 1.6, you get an editing frame, but it
;;;           won't let you type new values. This is true both when booting into OS9 and when running
;;;           in classic mode. In OSX native, you can edit.
;;; 7/16/2002 SVS Twist-downs work now, although they're not demonstrated very well in this file. See collection-databrowser
;;;           for a good demo of twist-downs.
;;; 7/18/2002 SVS now passing comparing-p to lookup-browser-data for more flexibility.
;;;           databrowser-compare-proc --> databrowser-item-compare-proc. More consistent.
;;; 10/15/2002 SVS browser's columns are now reorderable by dragging them (see draggable-columns).
;;; 1-29-2003 SVS combined basic-databrowser and collection-databrowser together in this file.
;;; Jan-Feb 2003 SVS many, many improvements.
;;; Friday December 5, 2003 SVS focus-ring attribute with help from GB.

#|
scroll wheel support under OSX (this was a real bear and it's still not done in the politically-correct manner)
lots of new introspection functions on columns, rows, scrollbars
simplified and cleaned up the API
key handler behavior so arrow keys etc. work
rightarrow twists a container open, left closes it
option-rightarrow twists a container open recursively
:WITH-SECONDS option so times can show seconds
employee-browser example
reveal-browser-row is working
editing cells API is working (even on Classic)
|#

(eval-when (:compile-toplevel :eval-toplevel :execute)
  (unless (find-package :browser)
    (defpackage "DATABROWSER"
      (:nicknames :brw :browser)
      (:export "CHILDREN"
               "CONTAINERP"
               "LOOKUP-BY-ROW"
               "LOOKUP-BY-OBJECT"
               "COLLECTION-DATABROWSER"
               "UPDATE"
               "REMOVE-ALL-BROWSER-ITEMS"
               "REVEAL-BROWSER-ROW"
               "REVEAL-BROWSER-DATA"
               "ADD-ITEMS"
               "ALLOW-EDITS"
               "SELECTED-ROWS"
               "DATABROWSER-STATE"
               "DATABROWSER-ITEM-ADDED"
               "DATABROWSER-ITEM-REMOVED"
               "DATABROWSER-EDIT-STARTED" 
               "DATABROWSER-EDIT-STOPPED"
               "DATABROWSER-ITEM-SELECTED"
               "DATABROWSER-ITEM-DESELECTED"
               "DATABROWSER-ITEM-DOUBLE-CLICKED"
               "DATABROWSER-CONTAINER-OPENED"
               "DATABROWSER-CONTAINER-CLOSING"
               "DATABROWSER-CONTAINER-CLOSED"
               "DATABROWSER-CONTAINER-SORTING"
               "DATABROWSER-CONTAINER-SORTED"
               "DATABROWSER-TARGET-CHANGED"
               "DATABROWSER-USER-STATE-CHANGED"
               "DATABROWSER-SELECTION-SET-CHANGED" 
               "DATABROWSER-USER-TOGGLED-CONTAINER")
               (:use "CL" "CCL"))))


; Put event-handler-patch.lisp in your Examples folder if this chokes.
(eval-when (:load-toplevel :execute)
  (unless (fboundp 'ccl::scroll-wheel-handler)
    (require "EVENT-HANDLER-PATCH")))

(in-package :databrowser)

(defparameter *log-browser-errors* t "True if you want errors that occur during browser callbacks logged.")

(defparameter *browser-log-stream* *error-output* "Where to log browser errors that occur during callbacks,
     as well as nonfatal messages if any. Set to nil for no logging whatsoever.")

(defparameter *debug* t "True if you want interactive debugging features of the databrowser enabled.")

; Improvements needed:
; Documentation
; Change employee-browser to use CLOS so automatic setf accessor mechanism will work.
; Drag callbacks
; Get discontiguous multiple selections working with command key. DONE.
; Move #'find-thing-with-string to here from file-browser and make it general-purpose.
; 
; Colored text in cells. See GrabBag example for how to use a custom drawing proc to do this:
;   http://developer.apple.com/samplecode/Sample_Code/Human_Interface_Toolbox/Mac_OS_High_Level_Toolbox/GrabBag.htm
; Italic and bold text in cells.
; Cmd-A to select all.
; Other :PROPERTY-TYPEs like :ICON, :PROGRESS-BARs, :POPUP, etc.
; Anything else the Carbon databrowser supports that we're not supporting now.
; Column view API (probably not until Apple documents it for Carbon).
; Add a self-identifying Row ID that optionally shows up in the table.

(require :appearance-manager "ccl:examples;appearance-manager-folder;appearance-manager.lisp")
(require :new-control-dialog-item)

#| Already loaded now. Now incorporated into controldefinitions.lisp and (reindex-interfaces) has been done
(require :controldefinitions-patch (make-pathname :directory 
                                                  (butlast
                                                   (pathname-directory
                                                    (or *compile-file-pathname*
                                                        *load-pathname*)))
                                                  :name "controldefinitions-patch"
                                                  :type "lisp"))

|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; DATABROWSER

(defclass ccl::auto-resizable-view-mixin ()
  ((ccl::previous-container-size :initarg :previous-container-size :initform nil
                            :accessor ccl::previous-container-size
                            :documentation "For use by auto-resize-view methods.")))

(defclass databrowser (key-handler-mixin ccl::new-control-dialog-item ccl::auto-resizable-view-mixin)
  (
   #| no longer needed
  (property-types :initarg :property-types :initform nil :accessor property-types
                   :documentation "An a-list of (property-id . keyword) where keyword is
                   one of :TEXT, :CHECKBOX, :TIME, etc.  Best I can tell, there's no
                   easy way to tell what property-type a property was defined with
                   after the fact, so we'll store it here.")
     |#
   (display-nil-as-blank :initarg :display-nil-as-blank :initform t :accessor display-nil-as-blank
                         :documentation "True if you want to display NIL values returned
                   from lookup-browser-data as blank strings. Generally, this is what
                   you want unless you're building some kind of Lisp inspector. If
                   false, NILs are displayed as the string 'NIL'.")
   (draggable-columns :initarg :draggable-columns :initform t :accessor draggable-columns
                      :documentation "True if you want the browser's columns to be reorderable by dragging.")
   (selection-type :initarg :selection-type :initform nil :accessor
                    selection-type
                    :documentation "One of (nil :single :contiguous :disjoint). nil is the default
                    case, which seems to be the same as :disjoint.")
   (allow-edits :initarg :allow-edits :initform t :accessor allow-edits
                :documentation "Global editing control disabler for the browser. If this is nil, editability
                                of all items in the browser is shut off, even for columns that are created to be editable.
                                Won't stop an edit if one is currently under way.")
   (vscrollp :initarg :vscrollp :initform t :accessor vscrollp
             :documentation "True if you want a vertical scrollbar initially.")
   (hscrollp :initarg :hscrollp :initform t :accessor hscrollp
             :documentation "True if you want a horizontal scrollbar initially.")
   (focus-ring :initarg :focus-ring :initform nil :accessor focus-ring
               :documentation "True if you want a focus ring to appear around the browser widget as a whole,
                  to indicate when it has focus. Usually this should be nil when the browser is the only widget
                  in the window.")
   ))

(defmethod finish-initializing ((me databrowser))
  (with-slots (selection-type vscrollp hscrollp) me
    (unless (memq selection-type '(nil :single :contiguous :disjoint))
      (ccl::report-bad-arg selection-type '(:single :contiguous :disjoint)))
    (enable-scroll-bars me vscrollp hscrollp) ; can do this anytime so we might as well do it here
    (let ((handle (dialog-item-handle me)))
      (case selection-type
        (:single (#_SetDataBrowserSelectionFlags handle #$kDataBrowserSelectOnlyOne))
        (:contiguous (#_SetDataBrowserSelectionFlags handle #$kDataBrowserNoDisjointSelection))
        
        ; Following seems to be the default case with the databrowser if you don't do anything.
        (:disjoint (#_SetDataBrowserSelectionFlags handle (+ #$kDataBrowserCmdTogglesSelection #$kDataBrowserDragSelect))))
      )))

;;; Stuff that could be added to MCL to provide a general mechanism for auto-resizing subviews.
;;; The idea is that any subview you want to be auto-resizable should inherit from auto-resizable-view-mixin.
;;; Then it contains a slot that has the previous size of its container, and you can write a method
;;; for it on auto-resize-view. 

(defun ccl::maintain-corner-relationship (old-position old-size new-size)
  "Find the corner of old-size that old-position is nearest and return
   a new-position that's the same distance away from the same corner in
   new-size."
  (let ((new-h (point-h old-position)) ; assume closer to topleft
        (new-v (point-v old-position))
        (temp 0))
    (if (> (point-h old-position) (setq temp (- (point-h old-size) (point-h old-position))))
      (setq new-h (- (point-h new-size) temp))) ; closer to right edge
    (if (> (point-v old-position) (setq temp (- (point-v old-size) (point-v old-position))))
      (setq new-v (- (point-v new-size) temp))) ; closer to bottom edge
    (make-point new-h new-v)))

(defmethod ccl::auto-resize-view ((view t) (container t))
  nil)

(defclass ccl::check-box-dialog-item-ar (check-box-dialog-item ccl::auto-resizable-view-mixin)
  ())

(defclass ccl::static-text-control-item-ar (ccl::static-text-control-item ccl::auto-resizable-view-mixin)
  ())

(defclass ccl::button-dialog-item-ar (button-dialog-item ccl::auto-resizable-view-mixin)
  ())

(defclass ccl::SCROLLING-FRED-VIEW-AR (scrolling-fred-view ccl::auto-resizable-view-mixin)
  ())

(defmethod ccl::auto-resize-view ((view ccl::check-box-dialog-item-ar) (container simple-view))
  (set-view-position view (ccl::maintain-corner-relationship (view-position view)
                                                     (ccl::previous-container-size view)
                                                     (view-size container))))

(defmethod ccl::auto-resize-view ((view ccl::button-dialog-item-ar) (container simple-view))
  (set-view-position view (ccl::maintain-corner-relationship (view-position view)
                                                     (ccl::previous-container-size view)
                                                     (view-size container))))

(defmethod ccl::auto-resize-view ((view ccl::static-text-control-item-ar) (container simple-view))
  (set-view-position view (ccl::maintain-corner-relationship (view-position view)
                                                     (ccl::previous-container-size view)
                                                     (view-size container))))

(defmethod ccl::auto-resize-view ((view ccl::SCROLLING-FRED-VIEW-AR) (container simple-view))
  (let ((delta (subtract-points (view-size container) (ccl::previous-container-size view))))
    (set-view-size view (add-points (view-size view) delta))))

(defmethod ccl::previous-container-size ((view t))
  nil)

(defmethod (setf ccl::previous-container-size) ((size t) (view t))
  nil)

; Specialize this for various types of subviews. Second parameter provides a means for further 
;   specialization as needed. You may need further specialization in cases where, e.g. you
;   have two databrowsers in a window, necessitating that they be both repositioned as well
;   as resized.
(defmethod ccl::auto-resize-view ((view databrowser) (container simple-view))
  (let ((delta (subtract-points (view-size container) (ccl::previous-container-size view))))
    (set-view-size view (add-points (view-size view) delta))))

; Following two methods should take care of all auto-resizing from now on, provided
;   methods for auto-resize-view are written.
(defmethod set-view-size :before ((view window) h &optional v)
  (declare (ignore h v))
  (let* ((old-size (view-size view)))
    (dovector (subview (view-subviews view))
      (setf (ccl::previous-container-size subview) old-size))))

(defmethod set-view-size :after ((view window) h &optional v)
  (declare (ignore h v))
  (dovector (subview (view-subviews view))
    (ccl::auto-resize-view subview view)))

;;; End of auto-resizing stuff

(defmethod view-key-event-handler ((browser databrowser) char &aux part)
  (when (ccl::installed-item-p browser)
    (with-focused-dialog-item (browser)
      (let ((handle (dialog-item-handle browser))
            (modifiers (ccl::get-modifiers)))
        (declare (dynamic-extent handle)
                 (type macptr handle))
        (when handle
          (setq part (#_HandleControlKey handle (char-code char) (char-code char) modifiers))) ; let the databrowser handle the key
        ))))

;--- BROWSER SCROLL BAR FUNCTIONS

(defmethod scroll-bars-enabled-p ((browser databrowser))
  "Returns 2 values (vertical and horizontal) indicating whether scroll bars are enabled for this browser.
   Note that this only reflects whether scroll bars are enabled or not--this will return T if a scroll bar
   is enabled but its thumb is invisible because it's not needed."
  (rlet ((h :boolean)
         (v :boolean))
    (#_GetDataBrowserHasScrollBars (dialog-item-handle browser) h v)
    (values (pref v :boolean)
            (pref h :boolean))))

(defmethod enable-scroll-bars ((browser databrowser) v-bool h-bool)
  "Give this two values (vertical and horizontal) indicating whether to enable scroll bars on this browser.
   This is dynamic so you can change it on the fly on an extant databrowser.
   This function WILL affect future calls to #'scroll-bars-enabled-p."
  (#_SetDataBrowserHasScrollBars (dialog-item-handle browser) h-bool v-bool))

(defmethod auto-size-columns ((browser databrowser))
  "Resize all columns to fill available horizontal space. Only works
   if horizontal scroll bar is disabled."
  (multiple-value-bind (v-enabled h-enabled) (scroll-bars-enabled-p browser)
    (declare (ignore v-enabled))
    (when (not h-enabled)
      (#_AutoSizeDataBrowserListViewColumns (dialog-item-handle browser)))))

(defmethod scroll-bars-active-p ((browser databrowser))
  "Here's how you can tell whether there's a scroll bar that's both enabled AND has a visible thumb.
   For each scroll bar, returns either NIL (meaning no active scroll bar) or an integer representing
     the maximum setting of that scroll bar."
  (multiple-value-bind (v-enabled h-enabled) (scroll-bars-enabled-p browser)
    (multiple-value-bind (v-max h-max) (get-scroll-max browser)
      ; These will be zero if scroll bar is enabled but has an invisible thumb
      (values (and v-enabled (not (zerop v-max)) v-max)
              (and h-enabled (not (zerop h-max)) h-max)))))
      
(defmethod get-scroll-position ((browser databrowser))
  "Returns current vertical and horizontal scrolling positions of the browser."
  (rlet ((top :unsigned-long)
         (left :unsigned-long))
    (#_GetDataBrowserScrollPosition (dialog-item-handle browser) top left)
    (values (pref top :unsigned-long) (pref left :unsigned-long))))

(defmethod set-scroll-position ((browser databrowser) top left)
    (#_SetDataBrowserScrollPosition (dialog-item-handle browser) top left))

(defmethod get-scroll-max ((browser databrowser))
  "Returns current vertical and horizontal scrolling maxima of the browser.
   (0 is always the minimum, so we don't provide a function called get-scroll-min.)
   Note that this may return a nonzero value even if scroll bar is not enabled, so you
   also have to check whether it's enabled to determine if this value is meaningful.

   Astonishingly, the databrowser API provides no mechanism for getting this, which
   means it's very difficult to support wheel mice.
   (If you don't pay attention to these limits, you can easily scroll the browser
   such that its content is about a mile above your screen. No error is given.)
   This is borne out by the
   fact that most of Apple's databrowser demo code doesn't support wheel mice. But
   of course the Finder supports wheel mice in its use of the databrowser, so there
   had to be a way. There is. This is it. Scalpel!"
  (let ((browser-handle (dialog-item-handle browser))
        (swap nil)
        (hmax 0)
        (vmax 0))
    (rlet ((SubControl :pointer))
      (flet ((getit (index)
               (rlet ((bounds :rect))
                 (#_GetIndexedSubControl browser-handle index SubControl)
                 (if (handlep (%get-ptr SubControl))
                   (prog1
                     (logand (#_GetControl32BitMaximum (%get-ptr SubControl))
                             #x7FFFFFFF) ; bit 31 is set. Damnifiknow why. It ain't negative.
                     (#_GetControlBounds (%get-ptr SubControl) bounds)
                     ;(format t "Index ~D: top=~D, left=~D, bottom=~D, right=~D~%"
                     ;        index 
                     ;        (pref bounds :rect.top)
                     ;        (pref bounds :rect.left)
                     ;        (pref bounds :rect.bottom)
                     ;        (pref bounds :rect.right))
                     
                     ; DAMN this is a silly, stupid, error-prone PITA! How come there's no
                     ;   easy way to figure out which scroll bar is the freakin' VERTICAL one?!?
                     (case index
                       (2 (when (> (- (pref bounds :rect.bottom) (pref bounds :rect.top))
                                   (- (pref bounds :rect.right)  (pref bounds :rect.left)))
                            (setf swap t)))
                       (1 (when (> (- (pref bounds :rect.right)  (pref bounds :rect.left))
                                   (- (pref bounds :rect.bottom) (pref bounds :rect.top)))
                            (setf swap t))))
                     )
                   (progn
                     (format t "Index ~D is bad" index)
                     0)) ; return 0 in pathological case. I've never seen this happen.
                 ; even if scrollbar thumbs are not there, system returns 0 anyway.
                 )))

        ; This makes little difference. There's often two valid control handles even
        ;   if one of the scrollbars is disabled.
        ;(multiple-value-bind (v-enabled h-enabled) (scroll-bars-enabled-p browser)

        (setf vmax (getit 1)) ; assume 1 is vertical until proven otherwise
        (setf hmax (getit 2))
        (if swap (rotatef vmax hmax))
        (values vmax hmax)))))

;;; Assuming you're using the version of event-handler-proc that calls scroll-wheel-handler, this provides
;;;   scroll wheel behavior for the databrowser.
;;; It's not as smooth as scrolling in Fred windows but that's probably because the databrowser
;;;   takes longer to refresh.

;;; There may be a way to just bypass all this and send the event directly to the databrowser,
;;;   where the event might be handled automatically, but I haven't found it yet.
;;;   And since most of Apple's own sample code for the databrowser doesn't support mouse wheel
;;;   scrolling, it seems a pretty good bet we'll have to do it manually.

(defmethod ccl::scroll-wheel-handler ((browser databrowser) delta direction wherep)
  (declare (ignore wherep))
  (if (> 10 (abs delta) 1) (setf delta (round delta 2))) ; heuristic to slow the wheel down a little
  (multiple-value-bind (venabled henabled) (scroll-bars-enabled-p browser)
    (flet ((update-thumb ()
             ; This sucks. Why does Apple make this necessary after setting scroll position?
             (when (or venabled henabled)
               (enable-scroll-bars browser nil nil)
               (enable-scroll-bars browser venabled henabled)
               ;(update browser t) ; this doesn't do the trick
               )))
      (when t ;(or venabled henabled) ; Hey! It seems pretty useful to be able to use the wheel
        ;  even when scrollbars are disabled! Let's leave it like this.
        (multiple-value-bind (vmax hmax) (get-scroll-max browser)
          (if (eq direction :vertical)
            (when (and ;venabled
                   (not (zerop vmax))) ; zero if enabled but thumb invisible
              (rlet ((rowheightptr :unsigned-word))
                (#_GetDataBrowserTableViewRowHeight (dialog-item-handle browser) rowheightptr)
                (let ((rowheight (pref rowheightptr :unsigned-word)))
                  (multiple-value-bind (vpos hpos) (get-scroll-position browser)
                    (set-scroll-position browser (min vmax (max 0 (- vpos (* rowheight delta))))
                                              hpos))
                  (update-thumb)
                  
                  )))
            ; direction is :horizontal (user used shift key with scroll wheel)
            (when (and ;henabled
                   (not (zerop hmax))) ; zero if enabled but thumb invisible
              (multiple-value-bind (vpos hpos) (get-scroll-position browser)
                (set-scroll-position browser vpos (min hmax (max 0 (- hpos (* (ccl::adjust-horizontal-wheel-speed browser delta)))))))
              (update-thumb)))))))
  #$noerr)

;--- BROWSER ROW AND COLUMN UTILITY FUNCTIONS

(defmethod get-column-count ((browser databrowser))
  "Returns the total number of columns in given browser."
  (rlet ((countptr :unsigned-long))
    (#_GetDataBrowserTableViewColumnCount (dialog-item-handle browser) countptr)
    (pref countptr :unsigned-long)))

; This doesn't seem to work right. I can't get it to tell me the actual number of displayed
;   rows, regardless of whether some of the rows are contained within others (twist-downs) or not. Seems flaky.
(defmethod get-row-count ((browser databrowser))
  "Returns the current total number of rows in given browser."
  (rlet ((countptr :unsigned-long))
    (#_GetDataBrowserItemCount (dialog-item-handle browser) #$kDataBrowserNoItem t 0 countptr)
    (pref countptr :unsigned-long)))

(defmethod get-column-ID ((browser databrowser) column-index)
  "Given a column index, where 0 is the current leftmost column, return the columnID for that column.
   Returns $kDataBrowserItemNoProperty if given index is out of range."
  (rlet ((columnIDptr :unsigned-long))
    (#_GetDataBrowserTableViewColumnProperty (dialog-item-handle browser) column-index columnIDptr)
    (pref columnIDptr :unsigned-long)))

(defmethod get-column-index ((browser databrowser) columnID)
  "Given a columnID return the column-index for that column. 0 is the column-index of the current leftmost column."
  (rlet ((indexptr :unsigned-long))
    (#_GetDataBrowserTableViewColumnPosition (dialog-item-handle browser) columnID indexptr)
    (pref indexptr :unsigned-long)))

(defmethod get-row-ID ((browser databrowser) row-index)
  "Given a row index, where 0 is the current topmost row, return the rowID for that row.
   Given that get-row-count is flaky w.r.t. twist-downs, this may be too."
  (rlet ((rowIDptr :unsigned-long))
    (#_GetDataBrowserTableViewItemID (dialog-item-handle browser) row-index rowIDptr)
    (pref rowIDptr :unsigned-long)))

(defmethod get-row-index ((browser databrowser) rowID)
  "Given a rowID return the row-index for that row. 0 is the row-index of the current topmost row.
   Given that get-row-count is flaky w.r.t. twist-downs, this may be too."
  (rlet ((indexptr :unsigned-long))
    (#_GetDataBrowserTableViewItemRow  (dialog-item-handle browser) rowID indexptr)
    (pref indexptr :unsigned-long)))


;--- MAIN LOWLEVEL API

; Specialize only if you invent a new kind of browser
(defmethod (setf lookup-browser-data) (data (browser databrowser) rowID columnID)
  "Specialize this."
  (format t "Setting ~S, ~S to ~S" rowID columnID data)
  data
  )

; Specialize only if you invent a new kind of browser
(defmethod lookup-browser-data ((browser databrowser) rowID columnID &optional (comparing-p nil))
  "Return a Lisp datum that is to be displayed at rowID, columID. Both
   will be 32-bit unsigned integers. Default
   method just returns a string '(rowID,columnID)'. Customize as needed for your application.
   Datum will automatically be converted to a string before display, if in fact
   a string is needed. If a checkbox is to be affected by this datum, the normal
   Lisp true and false values are appropriate, or return the special value :MIXED
   to ensure the checkbox is drawn in mixed style."
  (declare (ignore comparing-p))
  (format nil "(~D,~D)" rowID columnID))

(defmethod stringify-browser-data ((datum t))
  "Final step to ensure that datum is formatted as a string for display in browser, for
   cells that require strings. Customize if needed, but this should suffice for most purposes."
  (handler-case
    (string datum) ; deal with symbols quickly
    (error () (format nil "~A" datum))))

(defmethod stringify-browser-data ((datum string))
  "Deal with strings quickly."
  datum)

(defmethod stringify-browser-data ((datum integer))
  (format nil "~D" datum))

(defmethod %databrowser-item-data ((browser databrowser) rowID (property (eql #$kDataBrowserItemIsEditableProperty)) item-data set?)
  "Called by the databrowser to determine if the given rowID is editable. Behavior here works pretty well for text
   cells, but not for checkboxes."
  (declare (ignore rowID set?))
  (#_SetDataBrowserItemDataBooleanValue item-Data (allow-edits browser))
  ; assume for now that any element that has the #$kDataBrowserPropertyIsEditable property
  ;  is in fact editable. (Any element that does not have this property will never invoke this method.)
  ;  Setting true here just gives you an outline around the thing you want to edit. If you do
  ;  (#_SetDataBrowserItemDataBooleanValue item-Data nil) instead, you won't even get the outline.
  ; The allow-edits check above doesn't really need to be here with collection-databrowsers, because
  ;   the :after method on (setf allow-edits) prevents this method from ever being called.
  ;   Plus this doesn't prevent edits of checkboxes, at least visually, while the :after method does.
  ;   But of course, the :after method is per-column, while this method is per-row. So there's the tradeoff.
  nil)

(defmethod lookup-property-type ((browser databrowser) columnID)
  (declare (ignore columnID))
  "Only option for basic databrowser, which should never be instantiated anyway."
  :TEXT)

(defmethod set-sort-column ((browser databrowser) columnID)
  "Programmatically set the sort column for the browser. Most of the time, you should let the user select this
   by clicking on column headers."
  (#_SetDataBrowserSortProperty (dialog-item-handle browser) columnID))

(defmethod get-sort-column ((browser databrowser))
  "Retrieve the current columnID being sorted for the browser."
  (rlet ((ptr :unsigned-long))
    (#_GetDataBrowserSortProperty (dialog-item-handle browser) ptr)
    (pref ptr :unsigned-long)))

; Specialize only if absolutely necessary.
(defmethod %databrowser-item-data ((browser databrowser) rowID columnID item-data set?)
  "This method is called by the browser and should return a datum to be displayed where
   (columnID,rowID) is an (x,y) location. This function is only called for display, not
   for comparison. The returned datum should be a Lisp datum, but its specific Lisp type
   depends on the browser's property-type for this column."

  (when (> columnID 1023) ; because the rest are reserved by Apple
    (let ((property-type (lookup-property-type browser columnID)))
           ;(cdr (assoc columnID (property-types browser))))) ; yes this is slow.
                     ; but it's probably not worth speeding it up because:
                     ;   a. This routine (and thus the assoc call) is only used for displaying, not comparing.
                     ;   b. This routine is not called often, because the browser does a pretty
                     ;      good job of caching after the first call.
                     ;   c. Which is not the case when comparing. Comparing isn't cached, AFAIK.
                     ;      But see point a.
      (if set?
        (ecase property-type
          (:TEXT (let ((newstring))
                   (rlet ((theCFptr :pointer))
                     (#_GetDataBrowserItemDataText item-Data theCFptr)
                     (setf newstring (ccl::%get-cfstring (%get-ptr theCFptr)))
                     (#_CFRelease (%get-ptr theCFptr))
                     (setf (lookup-browser-data browser rowID columnID) newstring)
                     ))
                 )
          (:CHECKBOX ; grab checkbox value and stash it
           (rlet ((bval :ThemeButtonValue))
             (#_GetDataBrowserItemDataButtonValue item-Data bval)
             (let ((button-value (pref bval :ThemeButtonValue)))
               ;(format t "~%button-value = ~S" button-value)
               (when (allow-edits browser) ; this clause is really a kludge that's unnecessary with collection-databrowsers
                 (case button-value        ;   because there, the :after method on (setf allow-edits) prevents set? being true here
                   (#.#$kThemeButtonOn  (setf (lookup-browser-data browser rowID columnID) t))
                   (#.#$kThemeButtonOff (setf (lookup-browser-data browser rowID columnID) nil))
                   (t (setf (lookup-browser-data browser rowID columnID) :MIXED)))))))
          
          ; rest of these aren't really supported by this code yet
          (:POPUP #$kDataBrowserpopupMenuType) ; this could be changed by the user
          (:ICONANDTEXT #$kDataBrowserIconAndTextType) ; and I suppose this could too
          )
        
        (let ((application-data (lookup-browser-data browser rowID columnID))) ; ideally, this shouldn't care about property-type. It should just be Lisp data.
          (ecase property-type
            (:TEXT  (let ((string (if application-data
                                    (let ((string (stringify-browser-data application-data)))
                                      (if (zerop (length string))
                                        "    " ; 4 spaces (see below)
                                        string))
                                    (if (display-nil-as-blank browser)
                                      "    " ; 4 spaces so we'll have something to edit if editable.
                                             ; This is a kludge but it's extremely difficult to select the item
                                             ;   for editing if there isn't at least 1 character present
                                      (stringify-browser-data application-data) ; shows "NIL"
                                      ))
                                    )) ; ensure it's a string
                      (ccl::with-cfstrs ((theData string))
                        (#_SetDataBrowserItemDataText item-Data theData))))
            (:CHECKBOX (#_SetDataBrowserItemDataButtonValue item-Data (if application-data
                                                                        (if (eq application-data :MIXED) ; kludge. Need true, false, and mixed
                                                                          #$kThemeButtonMixed
                                                                          #$kThemeButtonOn)
                                                                        #$kThemeButtonOff)))
            (:TIME
             (rlet ((dt :LongDateCvt :lHigh 0 :lLow 0))
               (if (numberp application-data) ; if it's an integer, assume it's a Lisp universal-time. Otherwise just display current time.
                 ; If you want to display time in other than the browser's default display format, format
                 ;   it yourself in a :TEXT column
                 (unless (integerp application-data)
                   (setf application-data (truncate application-data)))
                 (setf application-data (get-universal-time)))
               (setf application-data (ccl::universal-to-mac-time application-data)) ; because Mac time is 1904-based and Lisp is 1900
               (setf application-data (truncate application-data))
               (let ((time-hi (logand #xFFFFFFFF00000000 application-data))
                     (time-lo (logand #xFFFFFFFF application-data)))
                 (%put-long dt time-hi 0)
                 (%put-long dt time-lo 4)
                 (#_SetDataBrowserItemDataLongDateTime item-Data dt)
                 ; or  (#_SetDataBrowserItemDataDateTime item-Data (get-time-unsigned-long) ; this is the Mac time. 1904 based.
                 ;    but this will roll over in 2036 or 2040, depending on whose base you use
                 )
               ))

            ; rest of these aren't really supported by this code yet
            (:ICON nil)
            (:PROGRESS-BAR nil)
            (:RELEVANCE nil)
            ; (:SLIDER nil) ; I think the documentation's wrong on this one. Seems to not be in Carbon.
            (:POPUP nil)
            (:ICONANDTEXT nil)
            )))
      nil
      ;;(#_SetDataBrowserItemDataDrawState item-Data #$kThemeStateActive) ; this doesn't help
      )))

(defmethod databrowser-compare ((browser databrowser) id1 id2 sortProperty)
  "Customize this if needed, but this method will usually suffice."
  (when (> sortProperty 1023) ; because the rest are reserved by Apple
    (let ((item1 (lookup-browser-data browser id1 sortProperty t))
          (item2 (lookup-browser-data browser id2 sortProperty t)))
      (labels ((compareit (item1 item2)
                 (cond ((and (stringp item1)
                             (stringp item2))
                        (string-lessp item1 item2))
                       ((and (numberp item1)
                             (numberp item2))
                        (< item1 item2))
                       ((null item1) t) ; consider nil less than true
                       ((null item2) nil)
                       (t ; as a last resort, compare the stringified versions of the data
                        (let ((sitem1 (stringify-browser-data item1))
                              (sitem2 (stringify-browser-data item2)))
                          (compareit sitem1 sitem2))))))
        (compareit item1 item2)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; CALLBACKS

(ccl::add-pascal-upp-alist 'databrowser-item-data-proc  #'(lambda (procptr) (#_NewDataBrowserItemDataUPP procptr)))
(ccl::add-pascal-upp-alist 'databrowser-Item-notification-proc  #'(lambda (procptr) (#_NewDataBrowserItemNotificationUPP procptr)))
(ccl::add-pascal-upp-alist 'databrowser-item-compare-proc    #'(lambda (procptr) (#_NewDataBrowserItemCompareUPP procptr)))
; add more here as needed like the drag handlers, contextual menu, and help callbacks

(defmethod log-browser-error (browser routine-name error &rest parms)
  "Deal with Lisp errors that occur during browser callbacks. Specialize if needed."
  (when *browser-log-stream*
    (progn (format *browser-log-stream* "Error in browser ~S, call = (~S~{ ~S~}).~%"
                   browser routine-name parms)
           (ccl::report-condition error *browser-log-stream*)
           nil)))

(defun maybe-log-browser-ref-error (browser-ref routine-name error &rest parms)
  "Low-level recorder for browser errors that occur during browser callbacks. Shouldn't need specialization."
     (when *log-browser-errors*
       (let ((browser (ignore-errors (handle.databrowser browser-ref))))
          (apply #'log-browser-error browser routine-name error parms))))

(defpascal databrowser-item-data-proc (:controlref browser 
                                                   :DataBrowserItemID item ; row
                                                   :DataBrowserPropertyID property ; column
                                                   :DataBrowserItemDataRef item-data
                                                   :boolean set-value
                                                   ;:without-interrupts nil ; this doesn't seem to do anything, t or nil
                                                   :osstatus) 
  (or (handler-case
        (%databrowser-item-data (handle.databrowser browser) item property item-data set-value)
        (error (c) (maybe-log-browser-ref-error browser '%databrowser-item-data c item property set-value)))
      #$noErr))

(defpascal databrowser-Item-notification-proc (:controlref browser
                                                           :DataBrowserItemID item
                                                           :DataBrowserItemNotification message
                                                           :void)
  (let ((bb (handle.databrowser browser))
        (fn nil))
    (setf fn 
          (case message
            (#.#$kDataBrowserItemAdded ;1          ;   The specified item has been added to the browser
             'databrowser-item-added)
            (#.#$kDataBrowserItemRemoved ;2          ;   The specified item has been removed from the browser
             'databrowser-item-removed)
            (#.#$kDataBrowserEditStarted ;3        ;   Starting an EditText session for specified item
             'databrowser-edit-started)
            (#.#$kDataBrowserEditStopped ;4        ;   Stopping an EditText session for specified item
             'databrowser-edit-stopped)
            (#.#$kDataBrowserItemSelected ;5       ;   Item has just been added to the selection set
             'databrowser-item-selected)
            (#.#$kDataBrowserItemDeselected ;6     ;   Item has just been removed from the selection set
             'databrowser-item-deselected)
            (#.#$kDataBrowserItemDoubleClicked ;7
             'databrowser-item-double-clicked)
            (#.#$kDataBrowserContainerOpened ;8    ;   Container is open  
             'databrowser-container-opened)
            (#.#$kDataBrowserContainerClosing ;9   ;   Container is about to close (and will real soon now, y'all)  
             'databrowser-container-closing)
            (#.#$kDataBrowserContainerClosed ;10   ;   Container is closed (y'all come back now!)  
             'databrowser-container-closed)
            (#.#$kDataBrowserContainerSorting ;11  ;   Container is about to be sorted (lock any volatile properties)  
             'databrowser-container-sorting)
            (#.#$kDataBrowserContainerSorted ;12   ;   Container has been sorted (you may release any property locks)  
             'databrowser-container-sorted)
            (#.#$kDataBrowserTargetChanged ;15     ;   The target has changed to the specified item  
             'databrowser-target-changed)
            (#.#$kDataBrowserUserStateChanged ;13  ;   The user has reformatted the view for the target  
             'databrowser-user-state-changed)
            (#.#$kDataBrowserSelectionSetChanged ;14;   The selection set has been modified (net result may be the same)  
             'databrowser-selection-set-changed)
            (#.#$kDataBrowserUserToggledContainer ; 16 ; _User_ requested container open/close state to be toggled
             'databrowser-user-toggled-container)
            (T (when *browser-log-stream*
                 (format *browser-log-stream* "~%Unknown message ~S for item ~S" message item))
               nil)))
    (when fn
     (handler-case
       (funcall fn bb item)
      (error (c) (maybe-log-browser-ref-error browser fn c bb item))))))

; This works--TN
(defpascal databrowser-item-compare-proc (:controlref browser
                                                 :DataBrowserItemID itemOne
                                                 :DataBrowserItemID itemTwo
                                                 :DataBrowserPropertyID sortProperty
                                                 :long)
  (if
    (handler-case
      (databrowser-compare (handle.databrowser browser) itemOne itemTwo sortProperty)
      (error (c) (maybe-log-browser-ref-error browser 'databrowser-compare c itemOne itemTwo sortProperty)))
    #xFFFFFFF
    0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; NOTIFICATION CALLBACK METHODS. BROWSER WILL CALL THESE WHEN EVENTS HAPPEN. SPECIALIZE AS NEEDED.

(defmethod databrowser-item-added            ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-item-removed          ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-edit-started          ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-edit-stopped          ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-item-selected         ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-item-deselected       ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-item-double-clicked   ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-container-opened      ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-container-closing     ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-container-closed      ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-container-sorting     ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-container-sorted      ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-target-changed        ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-user-state-changed    ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-selection-set-changed ((browser databrowser) rowID)
  (declare (ignore rowID)))
(defmethod databrowser-user-toggled-container ((browser databrowser) rowID)
  (declare (ignore rowID)))

(defmethod databrowser-state ((browser databrowser) rowID)
  "Call this to discover the current state of item at rowID."
  (rlet ((state :DataBrowserItemState))
    (#_GetDataBrowserItemState
     (dialog-item-handle browser)
     rowID
     state)
    (let ((result (pref state :DataBrowserItemState)))
      (if (= result #$kDataBrowserItemNoState)
        nil
        (values ; pretty self-explanatory
         (logtest #$kDataBrowserItemIsSelected result)
         (logtest #$kDataBrowserContainerIsOpen result)
         (logtest #$kDataBrowserItemIsDragTarget result))))))

(defmethod databrowser-items ((browser databrowser) state)
  "Returns a list of rowIDs for the rows that match state, followed by a count
   of items found (which should equal the length of the first list). State
   should be one of #$kDataBrowserItemIsSelected,  $kDataBrowserContainerIsOpen,
   or $kDataBrowserItemIsDragTarget.
   Note that if you frequently need just the count of items that match state, there's
   a faster way to do it (#_GetDataBrowserItemCount) that I haven't bothered to write an API for
   (at least, not a comprehensive one)."
  (let ((%discovered-items (#_NewHandle 0))
        (discovered-items nil)
        (discovered-count nil))
    (#_GetDataBrowserItems (dialog-item-handle browser)
     #$kDataBrowserNoItem
     t
     state
     %discovered-items)
    (setf discovered-count (truncate (#_GetHandleSize %discovered-items) #.(record-length :DataBrowserItemID)))
    (ccl::with-dereferenced-handle (ptr %discovered-items)
      (setf discovered-items (loop for i from 0 below discovered-count
                                 collect (%get-unsigned-long ; Inelegant, because this assumes
                                          ; #.(record-length :DataBrowserItemID)
                                          ; is 4. But that's probably a safe assumption.
                                          ptr (* i #.(record-length :DataBrowserItemID))))))
    (#_DisposeHandle %discovered-items)
    (values discovered-items discovered-count)))

(defmethod selected-rows ((browser databrowser))
  "Returns a list of rowIDs of the rows that are selected. Any rows that are owned
    by turned down containers are returned in the flattened list as well. Order of 
    returned rowIDs is sort of random. It's not necessarily top-to-bottom or bottom-
    to-top in the browser. The presence of selected items in turned-down containers
    seems to mess up the ordering."
  (multiple-value-bind (items count) (databrowser-items browser #$kDataBrowserItemIsSelected)
    (values (nreverse items) count)))

(defmethod (setf selected-rows) (rowids (browser databrowser))
  "Sets selected rows to be those in the rowids list or array. Any others
   currently selected will be deselected."
  (let ((offset 0)
        (len (length rowids)))
    (unless (zerop len)
      (%stack-block ((ids (* 4 len) :clear t))
        (map nil #'(lambda (rowid) (%put-long ids rowid offset) (incf offset 4)) rowids)
        (ccl::errchk
         (#_SetDataBrowserSelectedItems (dialog-item-handle browser) len ids #$kDataBrowserItemsAssign))))))

(defmethod first-selected-row ((browser databrowser))
  "Emulate behavior from table-dialog-items. Except this returns
   row IDs rather than points representing cells."
  (first (selected-rows browser)))

(defmethod window-can-do-operation ((browser databrowser) op &optional item)
  (declare (ignore op item))
  nil)

(defmethod update ((browser databrowser) rowID)
  "Redraw given row. Pass t for rowID to redraw all rows."
  (if (eq rowID t)
    (setf rowID nil))
  (rlet ((items :unsigned-long))
    (%put-long items (or rowID #$kDataBrowserNoItem))
    (#_UpdateDataBrowserItems 
     (dialog-item-handle browser)
     #$kDataBrowserNoItem ; container ?? Do I really have to care about this?? Seems like not.
     1 ; just deal with one item (or all of them) at a time for now
     items
     #$kDataBrowserNoItem ; wimp out on the pre-sort. Make the browser do the work.
     #$kDataBrowserNoItem ; all properties. (Could do just one, but we didn't make an API for that).
     )))

(defmethod ccl::control-dialog-item-procid ((item databrowser))
  "Dummy for databrowser, but it has to be something. Nil won't work."
  8)

(defmethod install-view-in-window :after ((me databrowser) dialog)
  ; I don't like the fact that the browser's columns are made draggable by setting a property
  ;   of the window the browser is in, but whaddya gonna do?
  (#_SetAutomaticControlDragTrackingEnabledForWindow (wptr dialog) (draggable-columns me))
  (unless (osx-p) ; not necessary in OSX, but necessary in OS9 and classic (thanks Terje!)
    (with-rgb (forecolor *black-color*)
      (with-rgb (backcolor *white-color*)
        (rlet ((desc :ControlFontStyleRec
                     :font #$kControlFontViewSystemFont
                     :flags (logior #$kControlUseFontMask ;#$kControlUseModeMask #$kControlUseAllMask #$kControlAddToMetaFontMask
                                    #$kControlUseFaceMask #$kControlUseJustMask)
                     ;(logior #$kControlAddToMetaFontMask ; without this, you can't apply italics or boldface to #$kControlFontViewSystemFont
                     ;        #$kControlUseAllMask 
                     ;        )
                     :just #$teCenter
                     :style 0
                     :ForeColor forecolor
                     :BackColor backcolor))
          (#_SetControlFontStyle (dialog-item-handle me) desc)))))
  ; (add-key-handler me dialog) ; needed if databrowser does not inherit from key-handler-mixin
  (finish-initializing me) ; can't be done until browser is in a window
  )

#| Methods needed if databrowser does not inherit from key-handler-mixin
(defmethod key-handler-p ((item databrowser))
  t)

(defmethod set-selection-range ((item databrowser) &optional pos curpos)
  (declare (ignore pos curpos))
  0)

(defmethod enter-key-handler ((item databrowser) last-key-handler)
  (declare (ignore last-key-handler))
  nil)

(defmethod exit-key-handler ((item databrowser) next-key-handler)
  (declare (ignore next-key-handler))
  t)
|#

(defmethod allow-returns-p ((item databrowser))
  "This must be true to allow editing items in the browser after hitting <return>."
  t)

(defmethod allow-tabs-p ((item databrowser))
  nil)

(defmethod ccl::make-control-handle ((control databrowser) wptr)
  (ccl::with-item-rect (rect control)
    (rlet ((hhandle (:pointer (:handle :controlrecord)))
           (callbacks :DataBrowserCallbacks))
      (ccl::errchk (#_CreateDataBrowserControl ;; ## remember to dispose of this after use!
                    wptr
                    rect 
                    #$kDataBrowserListView
                    hhandle))
      (let ((handle (%get-ptr hhandle)))
        (setf (rref callbacks :DataBrowserCallbacks.version) #$kDataBrowserLatestCallbacks)
        (ccl::errchk (#_InitDataBrowserCallbacks callbacks))
        (setf (rref callbacks :DataBrowserCallbacks.itemDataCallback) ; this should be the correct syntax for the variant in Lisp
              databrowser-item-data-proc)
        ; (#_NewDataBrowserItemDataUPP MyGetSetItemData))
        (setf (rref callbacks :DataBrowserCallbacks.itemNotificationCallback)
              databrowser-Item-notification-proc)
        
        (setf (rref callbacks :DataBrowserCallbacks.itemCompareCallback)
              databrowser-item-compare-proc)
        
        (ccl::errchk (#_SetDataBrowserCallbacks handle callbacks))
        (setf (handle.databrowser handle)
              control)
        (setf (dialog-item-handle control)
              handle)
        ;(ccl::errchk (#_SetDataBrowserTarget handle 502)) ; this doesn't do anything, but they do it in the C version
        (ccl::errchk (#_SetDataBrowserActiveItems handle t)) ; doesn't seem to be necessary
        
        (rlet ((ptr :boolean (focus-ring control)))
          (ccl::errchk (#_SetControlData handle #$kControlNoPart #$kControlDataBrowserIncludesFrameAndFocusTag 1 ptr))
          ;                                                                 1! Whoda thunk it?          -------^
          )
                 
        (#_SetDataBrowserTableViewHiliteStyle handle
         #$kDataBrowserTableViewFillHilite
         ;#$kDataBrowserTableViewMinimalHilite  ; just highlights individual cells, not whole rows
         )
        ))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; MAP BETWEEN DATABROWSER HANDLE AND OBJECT

(defvar *databrowsers* (make-hash-table :test 'eql :weak :value))

(defun (setf handle.databrowser) (databrowser handle)
;  (ccl::errchk (#_SetControlProperty handle 1 1 4 databrowser))
  (setf (gethash handle *databrowsers*) databrowser))

(defun handle.databrowser (handle)
  (gethash handle *databrowsers*))

;; COLUMNS

(defvar %last-property-id% 1025)

(defun get-unused-property-id ()
  ;; ## need to be improved to safeguard against too large ids!
  (incf %last-property-id%))

; Note: #_GetDataBrowserItemDataProperty seems to return property flags if you need them later.

; make this pay attention to a full font-spec eventually
(defmethod databrowser-add-listview-column ((browser databrowser) &key property-id
                                                  (property-type :TEXT)
                                                  (fontstyle :normal)
                                                  (fontsize 12)
                                                  (minwidth 10) ; make minwidth and maxwidth different if you want columns resizable
                                                  (maxwidth 200)
                                                  (title "Untitled")
                                                  (with-seconds nil) ; true if you want :TIME property-type to display seconds
                                                  (justification :center)
                                                  (order :increasing)
                                                  (editable nil) ; true if items in this column can be edited
                                                  (position #$kDataBrowserListViewAppendColumn))
  (setf property-id (or property-id (get-unused-property-id)))
  ;;(setf (property-types browser) (acons property-id property-type (property-types browser))) ; keep track of what kind of property this was
  (let ((flags ;;#$kDataBrowserDefaultPropertyFlags)) ; don't use this one. Doesn't work.
         (+ #$kDataBrowserListViewDefaultColumnFlags #$kDataBrowserListViewSelectionColumn))
        (handle (dialog-item-handle browser)))
    (ccl::with-cfstrs ((titlestring title))
      (with-rgb (forecolor *red-color*) ; do-nothing. It's always black regardless of what you put here.
        (with-rgb (backcolor *white-color*)
          (rlet ((desc :DataBrowserListViewColumnDesc
                       :propertydesc.propertyId property-id
                       :propertydesc.propertyType
                       (ecase property-type
                         (:TEXT #$kDataBrowserTextType)
                         (:CHECKBOX #$kDataBrowserCheckboxType)
                         (:TIME #$kDataBrowserDateTimeType) ; we'll leave the flags alone, no provision for relative date/time herein
                         ; rest of these aren't really supported by this code yet
                         (:ICON #$kDataBrowserIconType)
                         (:PROGRESS-BAR #$kDataBrowserProgressBarType)
                         (:RELEVANCE #$kDataBrowserRelevanceRankType)
                         ; (:SLIDER #$kDataBrowserSliderType) ; I think the documentation's wrong on this one. Seems to not be in Carbon.
                         (:POPUP #$kDataBrowserpopupMenuType)
                         (:ICONANDTEXT #$kDataBrowserIconAndTextType)
                         )
                       :propertydesc.propertyFlags (logior flags
                                                           (if editable #$kDataBrowserPropertyIsEditable 0)
                                                           (if (and (eq property-type :TIME)
                                                                    with-seconds) #$kDataBrowserDateTimeSecondsToo 0))
                       :headerbtndesc.version #$kDataBrowserListViewLatestHeaderDesc 
                       :headerbtndesc.initialorder (ecase order
                                                     ;((NIL) #$kDataBrowserOrderUndefined) ; Not supported. Assume Increasing.
                                                     (:increasing #$kDataBrowserOrderIncreasing)
                                                     (:decreasing #$kDataBrowserOrderDecreasing))
                       :headerbtndesc.minimumWidth (or minwidth maxwidth 100)
                       :headerbtndesc.maximumWidth (or maxwidth minwidth 100)
                       :headerbtndesc.btnContentInfo.contentType #$kControlContentTextOnly ; for now, no icons in headers
                       :headerbtndesc.btnFontStyle.font #$kControlFontViewSystemFont
                       :headerBtnDesc.btnFontStyle.flags (logior #$kControlAddToMetaFontMask ; without this, you can't apply italics or boldface to #$kControlFontViewSystemFont
                                                                 ; #$kControlAddFontSizeMask
                                                                 ; #$kControlUseAllMask ; TN says this is bad in OS9
                                                                 (logior #$kControlUseFontMask #$kControlUseJustMask #$kControlUseFaceMask)
                                                                 )
                       :headerBtnDesc.btnFontStyle.just (case justification
                                                          (:center #$teCenter)
                                                          (:left #$teFlushLeft)
                                                          (:right #$teFlushRight))
                       :headerbtndesc.btnFontStyle.style
                       (ecase fontstyle
                         (:normal 0)
                         (:bold 1)
                         (:italic 2)
                         (:bold-italic 3))
                       
                       :headerbtndesc.btnFontStyle.size fontsize
                       :headerbtndesc.btnFontStyle.mode #$srcOr
                       
                       :headerbtndesc.btnFontStyle.forecolor forecolor
                       
                       :headerbtndesc.btnFontStyle.backColor backcolor
                       
                       :headerbtndesc.titleOffset 0
                       :headerBtnDesc.titleString titlestring
                       ))
            (ccl::errchk (#_AddDataBrowserListViewColumn handle desc position))
            ))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric add-items (databrowser items &optional parent sortproperty)
  (:documentation "Pass a sequence of items in items parameter. If parent is given, it must
   be the dataID of an item of the same type as items, and which will be the parent of the list of items.
   That is, parent should be an item that has a twist-down triangle next to it and this routine
   would be called as a result of the user twisting it down. Normally, the programmer will never call this
   directly with parentID non-nil; that call typically only happens automatically when the user
   twists down a triangle."))

(defmethod add-items ((browser databrowser) items &optional parent sortproperty)
  "A silly do-nothing method since databrowser is not really designed for direct instantiation."
  (declare (ignore items parent sortproperty))
  (ccl::errchk (#_AddDataBrowserItems
           (dialog-item-handle browser)
           #$kDataBrowserNoItem ; no container (twist-down) in this demo
           5 ; 5 items
           (%null-ptr) ; let it generate its own DataIDs
           #$kDataBrowserItemNoProperty)))

(defgeneric remove-all-browser-items (databrowser &optional propertyID sortorder)
  (:documentation "Removes all rows from the databrowser. Call this if you need to completely change the list
   of items displayed. You can also pass propertyID and sortorder to pre-set the browser
   if you're about to add a list of pre-sorted items."))

(defmethod remove-all-browser-items ((browser databrowser) &optional propertyID sortorder)
  (#_RemoveDataBrowserItems (dialog-item-handle browser) #$kDataBrowserNoItem 0 (%null-ptr) #$kDataBrowserItemNoProperty)
  (when (and propertyID sortorder) ; let caller pre-set column to be sorted when items are added again
    (#_SetDataBrowserSortProperty (dialog-item-handle browser) propertyID)
    (#_SetDataBrowserSortOrder (dialog-item-handle browser)
     (case sortorder
       (:increasing #$kDataBrowserOrderIncreasing)
       (t #$kDataBrowserOrderDecreasing)))
     ))

(defmethod reveal-browser-data ((browser databrowser) rowID columnID &optional (also-select? t))
  "Reveal the given rowID, columnID in the browser. By default, also selects it."
  (#_RevealDataBrowserItem (dialog-item-handle browser) rowID columnID
   (if also-select? #$kDataBrowserRevealAndCenterInView #$kDataBrowserRevealWithoutSelecting))
    ;#$kDataBrowserRevealOnly is available too, but doesn't seem too useful
  )

(defmethod reveal-browser-row ((browser databrowser) rowID &optional (also-select? t))
  "Reveal the given rowID in the browser. By default, also selects it."
  (when rowID
    (reveal-browser-data browser rowID #$kDataBrowserItemNoProperty also-select?)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Collection-databrowser. The main high-level API.

(defvar *first-row-id* 2 "ID of first row in browser. Zero is not allowed.")
(defvar *first-column-id* 1025 "ID of first column in browser. First 1024 are reserved by Apple.")

(eval-when (:load-toplevel :compile-toplevel :execute)
  (setf *super-optimize-primary-slot-access* t)) ; best speed in MCL

(defclass two-way-table ()
  ((forward-table :initarg :forward-table :initform (make-hash-table) :accessor forward-table)
   (reverse-table :initarg :reverse-table :initform (make-hash-table :test 'equal) :accessor reverse-table)
   (next-id :initarg :next-id :initform *first-row-id* :accessor next-id))
  (:primary-p t) ; for best speed in MCL
  )

(defmethod clear-two-way-table ((table two-way-table))
  "Clear the table and reset its identifier to beginning."
  (clrhash (forward-table table))
  (clrhash (reverse-table table))
  (setf (next-id table) *first-row-id*))

(defclass collection-databrowser (databrowser)
  ((column-descriptors :initarg :column-descriptors :initform nil :accessor column-descriptors
                       :documentation "A sequence (use a vector for best speed) of column descriptors. See documentation.")
   (first-column-id :initarg :first-column-id :initform *first-column-id* :accessor
                     first-column-id
                     :documentation "ID of leftmost column")
   (object-table :initarg :object-table :initform (make-instance 'two-way-table) :accessor object-table)
   (triangle-space :initarg :triangle-space :initform t :accessor triangle-space
              :documentation "True if you want to allocate space for disclosure triangles in leftmost column.
                              (For now, the leftmost column is the only place they can go, although the underlying
                               API allows them to go in any column.)
                              You still have to ensure that #'containerp returns true for your row before
                              a triangle will actually appear."))
  (:documentation "Databrowser specialized for displaying a collection of objects, one per row, where
    columns represent attributes of the objects."))

; This is necessary so you can use arrow keys to move around in databrowser, when there is some
;   other kind of view in the same window that can also be a key handler. Without this method,
;   key control doesn't revert to the browser when you click in it.
(defmethod view-click-event-handler :after ((item databrowser) where)
  (declare (ignore where))
  (let ((my-dialog (view-window item)))
    (when my-dialog
      (if (neq item (current-key-handler my-dialog))
        (set-current-key-handler my-dialog item nil)))))

(defmethod lookup-property-type ((browser collection-databrowser) columnID)
  (with-slots (column-descriptors) browser
    (let ((column-descriptor (elt column-descriptors (- columnID (first-column-id browser)))))
      (column-property-type column-descriptor))))

(defmethod (setf allow-edits) :after (value (browser collection-databrowser))
  "Globally turn off/on editability of everything in the browser. Can turn off editability
   of columns defined to be editable, but doesn't turn on editability of columns defined not
   to be."
  ; Use this code as a guide for selectively turning off editability of specific columns if needed later.
  (let ((oldflags nil)
        (handle (dialog-item-handle browser)))
    (map nil #'(lambda (column-descriptor)
                 (let ((propertyID (property-ID column-descriptor)))
                   (rlet ((oldflagsptr  :unsigned-long))
                     (#_GetDataBrowserPropertyFlags handle propertyID oldflagsptr)
                     (setf oldflags (pref oldflagsptr :unsigned-long)))
                   (if value
                     (when (column-editable column-descriptor) ; only turn on columns that were originally defined to be editable
                       (#_SetDataBrowserPropertyFlags handle propertyID (logior oldflags #$kDataBrowserPropertyIsEditable)))
                     (#_SetDataBrowserPropertyFlags handle propertyID (logand oldflags (lognot #$kDataBrowserPropertyIsEditable)))
                     )))
         (column-descriptors browser))))

(defmethod window-can-do-operation ((browser collection-databrowser) op &optional item)
  (declare (ignore item))
  (and (eq op 'copy) (selected-rows browser)))

(defmethod copy ((browser collection-databrowser))
  "Default method just grabs a list of the objects in the selected rows."
  (let ((rows (selected-rows browser)))
    (when rows
      (put-scrap :LISP 
                 (mapcar #'(lambda (rowid) (lookup-by-row browser rowid))
                          rows)))))

(defmethod add-items ((browser collection-databrowser) items &optional parentID sortPropertyID)
  (let ((offset 0)
        (len (length items))
        )
    (unless (zerop len)
      (with-slots (forward-table reverse-table) (object-table browser)
        (%stack-block ((ids (* 4 len) :clear t)) ; God help us all
          (flet ((add-item (item)
                   (let ((numeric-id (gethash item reverse-table)))
                     (unless numeric-id ; maintain numeric-ids for items even if triangles open, close, and open again
                       (setf numeric-id (incf (next-id (object-table browser))))
                       (setf (gethash item reverse-table) numeric-id)
                       (setf (gethash numeric-id forward-table) item))
                     (%put-long ids numeric-id offset)
                     (incf offset 4))))
            (map nil #'add-item items)
            (ccl::errchk (#_AddDataBrowserItems
                          (dialog-item-handle browser)
                          (or parentID #$kDataBrowserNoItem)
                          len
                          ids
                          (or sortPropertyID #$kDataBrowserItemNoProperty)))))))))

(defmethod remove-all-browser-items :before ((browser collection-databrowser) &optional propertyID sortorder)
  (declare (ignore propertyID sortorder))
  (clear-two-way-table (object-table browser)))

(defclass column-descriptor ()
  ((function :initarg :function :initform nil :accessor column-function
             :documentation "Function (lambda) or symbolic name of a function to run against
             the row object to return value for this column.
             This should [usually] be a function of one parameter:
             the item in that row. Unless more-function-parameters is true, in which
             case 3 additional parameters are passed for a total of 4.")
   (editable :initarg :editable :initform nil :accessor column-editable
             :documentation "True if this column is editable. This attribute is overridable globally by
             the allow-edits slot of the databrowser.")
   (setter :initarg :setter :initform nil :accessor column-setter
           :Documentation "Function used to set a value after value was edited in an editable databrowser cell.
           [setter is ignored for non-editable columns because it isn't needed].
           Called with 2 args: (newvalue item). If setter is non-nil, and this column is editable, and
           the user edits a value, then when the user is done editing the value the setter will be called.
           It will be given newvalue which is the value the user entered. If setter is not provided in
           the descriptor-specification, we'll try to find a function named (setf column-function) and use that
           for the setter, where column-function is of course the slot immediately preceding this one.")
   (more-function-parameters :initarg :more-function-parameters :initform nil :accessor more-function-parameters
                             :documentation "True if the given function (the function slot) is a function of 4 parameters,
                   (item comparing-p browser rowID), rather than just the usual one.
                   First is just as described in the documentation for the function slot.
                   comparing-p will be true if the function is being called for comparison purposes;
                      false if it's just being called for display purposes. Thus this gives you a way
                      to return a different value depending on what the value will be used for.
                   browser is the current browser
                   rowID is the current rowID. 

                   The last two parameters give you the ability to make the appearance of the 
                   item dependent on the browser's visual state. Sometimes you care whether the
                   element is currently selected or whether it's a container that's opened, for
                   example.")
   (title :initarg :title :initform nil :accessor column-title)
   (property-id :initarg :property-id :initform nil :accessor property-id
                 :documentation "Used internally by the databrowser")
   (property-type :initarg :property-type :initform :TEXT :accessor column-property-type)
   (other-attributes :initarg :other-attributes :initform nil :accessor column-other-attributes))
  (:primary-p t) ; for best speed in MCL
  )

(defun parse-column-descriptor (descriptor-specification)
  (let ((descriptor (make-instance 'column-descriptor))
        (ptype nil))
    (cond ((consp descriptor-specification)
           (setf (column-function descriptor) (car descriptor-specification))
           (setf (more-function-parameters  descriptor) (getf (cdr descriptor-specification) :MORE-FUNCTION-PARAMETERS))
           (setf (column-title    descriptor) (or (getf (cdr descriptor-specification)   :TITLE)
                                                  (ignore-errors ; in case it's a lambda, car won't be stringable
                                                   (string (car descriptor-specification)))))
           (when (setf ptype (getf (cdr descriptor-specification) :PROPERTY-TYPE))
             (setf (column-property-type descriptor) ptype)) ; defaults to :TEXT
           (setf (column-editable     descriptor) (getf (cdr descriptor-specification) :EDITABLE))
           (setf (column-setter        descriptor) (getf (cdr descriptor-specification) :SETTER))
           (setf (column-other-attributes descriptor)
                 (ccl::remove-keywords (cdr descriptor-specification)
                                  '(:TITLE :EDITABLE :MORE-FUNCTION-PARAMETERS :PROPERTY-TYPE :FUNCTION :SETTER))))
          ((symbolp descriptor-specification)
           (setf (column-function descriptor) descriptor-specification)
           (setf (column-title    descriptor) (string descriptor-specification)))
          ((functionp descriptor-specification)
           (setf (column-function descriptor) descriptor-specification)
           (setf (column-title    descriptor) "Untitled"))
          (t (error "Invalid column descriptor specification ~S" descriptor-specification)))
     descriptor))

(defmethod shared-initialize :after ((descriptor column-descriptor) slot-names &rest initargs)
  (declare (ignore slot-names initargs))
  (let ((fname nil))
    (when (column-editable descriptor)
      (when (and (not (column-setter descriptor))
                 (setf fname (function-name (column-function descriptor))))
        (when (ignore-errors (fdefinition (list 'setf fname)))
          (setf (column-setter descriptor) (list 'setf fname)))))))

(defmethod finish-initializing ((browser collection-databrowser))
  "Try to deduce a setter function if we weren't given one. Note that this won't work for 'macro-style'
    setf forms like (setf car) and friends. It only works for user-defined setf functions and CLOS accessor
    methods."
  (call-next-method)
  (let ((columnID (first-column-id browser))
        (descriptor nil))
    (with-slots (column-descriptors) browser
      (setf column-descriptors
            (map 'vector ; might as well always use a vector
                 #'parse-column-descriptor
                 column-descriptors))
      
      (dotimes (i (length column-descriptors))
        (setf descriptor (elt column-descriptors i))
        (setf (property-id descriptor) columnID) ; for future reference
        (apply 'databrowser-add-listview-column
         browser
         :property-id columnID
         :property-type (column-property-type descriptor)
         :title (column-title descriptor)
         :editable (column-editable descriptor)
         (column-other-attributes descriptor)
         )
        (incf columnID))
      (when (and (triangle-space browser)
                 (> (length column-descriptors) 0))
        (#_SetDataBrowserListViewDisclosureColumn (dialog-item-handle browser) (first-column-id browser)  nil)
        )
      )))

; Specialize these only if you invent a new kind of browser
(defmethod lookup-by-row ((browser collection-databrowser) rowID)
  "Returns the object on the given rowID."
  (gethash rowID (forward-table (object-table browser))))

(defmethod lookup-by-object ((browser collection-databrowser) object)
  "Returns the rowID for an object given an object in the browser."
  (gethash object (reverse-table (object-table browser))))

(defmethod %databrowser-item-data ((browser collection-databrowser) rowID (property (eql #$kDataBrowserItemIsContainerProperty)) item-data set?)
  "This method is called automatically by the browser when it wants to know if an item is a container."
  (declare (ignore set?))
  (#_SetDataBrowserItemDataBooleanValue item-Data (containerp (lookup-by-row browser rowID)))
  nil)

(defmethod lookup-browser-data ((browser collection-databrowser) rowID columnID &optional (comparing-p nil))
  (with-slots (column-descriptors) browser
    (let ((rowitem (lookup-by-row browser rowID))
          (column-descriptor (elt column-descriptors (- columnID (first-column-id browser)))))
      (ignore-errors
       (if (more-function-parameters column-descriptor)
         (funcall (column-function column-descriptor) rowitem comparing-p browser rowID)
         (funcall (column-function column-descriptor) rowitem))))))

; Specialize only if you invent a new kind of browser
(defmethod (setf lookup-browser-data) (data (browser collection-databrowser) rowID columnID)
  (with-slots (column-descriptors) browser
    (let* ((rowitem (lookup-by-row browser rowID))
           (column-descriptor (elt column-descriptors (- columnID (first-column-id browser))))
           (column-setter (when column-descriptor (column-setter column-descriptor))))
      (when column-setter ; no edits can be permanent without this
        (funcall column-setter data rowitem)))))

; Specialize for your application
(defmethod databrowser-item-selected         ((browser collection-databrowser) rowID)
  (declare (ignore rowID))
  ; (update browser rowID) ; uncomment this if you have column functions that
  ;    change what they display based on selection status
  ;(format t "~%You selected ~S" (lookup-by-row browser rowID))
  )

; This one is handy for debugging. Gotta do cmd-option in actuality right now.
;   Don't use control because we'll eventually use that for contextual menus.
;   Can't use just command because, well, it doesn't work. It tries to make
;    a disjoint selection.
(defmethod databrowser-item-selected :around ((browser collection-databrowser) rowID)
  (when *debug*
    (let ((object (lookup-by-row browser rowID)))
      (when (and (option-key-p) (command-key-p)) ; quick way to inspect the object
        (inspect object))))
  (call-next-method))

; Specialize for your application if needed.
(defmethod databrowser-item-deselected         ((browser collection-databrowser) rowID)
  (declare (ignore rowID))
  ; (update browser rowID) ; uncomment this if you have column functions that
  ;    change what they display based on selection status
  ;(format t "~%You deselected ~S" (lookup-by-row browser rowID))
  )

; Specialize for your application
(defmethod containerp ((item t))
  "Specialize for your application. Item should be one of your row objects.
   Return true if you want this object to appear with a twist-down triangle."
  nil)

; Specialize for your application
(defmethod children ((item t))
  "Specialize for your application. Should return a sequence of subitems for this item,
   or just nil if there aren't any. It only makes sense for this to return non-nil
   if (containerp item) is also true."
  nil)

; Specialize for your application only if default method is insufficient.
(defmethod databrowser-container-opened ((browser collection-databrowser) rowID)
  "Called when user twists down a container's triangle.
   This default method should work for most purposes. Specialize if necessary, 
   but you should mostly specialize #'children."
  ;(format t "Opened row ~D" rowID)
  (let ((parent-thing (lookup-by-row browser rowID)))
    (add-items
     browser
     (children parent-thing)
     rowID)
    (update browser rowID) ; only necessary to assist column functions that
        ; change the displayed string depending on whether triangle is open or closed.
        ; These are always column functions that are specified with :more-function-parameters.
        ; Anyway, these functions will have no effect unless we call update
        ;   here, because the browser caches its cell values aggressively.
    ))

; Probably no need to specialize this.
(defmethod databrowser-container-closing ((browser collection-databrowser) rowID)
  "Apparently you don't really have to do anything here to cause proper display,
   except call update.
   And I'm only doing _that_ because it's necessary in the few cases where 
   some column function displays one value when a triangle is open vs. when it's closed.
   (See documentation for databrowser-container-opened.) 
   Customize for side-effects if needed."
  ;(format t "Closing row ~D" rowID)
  (update browser rowID)
  )

(eval-when (:load-toplevel :compile-toplevel :execute) 
  (pushnew (make-pathname :host (pathname-host *load-truename*)
                          :directory (append (pathname-directory *load-truename*)
                                             '("databrowser-examples")))
           ccl::*module-search-path*)) ; so require <databrowser-example> will work

(provide "DATABROWSER")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; EXAMPLE. See better examples in individual example files like
; file-browser.lisp
; search-files-browser.lisp
; employee-browser.lisp
; etc.
#|
;====Example 1: Basic raw browser. This uses the lowest-level API which should never be used in real programs.
(defun test-raw-browser ()
  (let ((databrowser (MAKE-INSTANCE
                       'databrowser
                       :allow-edits t ; change this after browser is created to see the effect
                       :VIEW-SIZE
                       #@(341 224)
                       :VIEW-POSITION
                       #@(91 58))))
    (MAKE-INSTANCE
      'COLOR-DIALOG
      :WINDOW-TYPE
      :DOCUMENT-WITH-GROW
      :VIEW-POSITION
      #@(105 74)
      :VIEW-SIZE
      #@(536 355)
      :back-color ccl::*green-color*
      :VIEW-SUBVIEWS
      (LIST (MAKE-DIALOG-ITEM 'ccl::STATIC-TEXT-DIALOG-ITEM #@(337 300) #@(54 16) "Example" 'NIL)
            databrowser))
  (databrowser-add-listview-column databrowser :fontstyle :bold :editable t)
  (add-items databrowser nil)
  databrowser
  ))

(test-raw-browser)

|#

