;; -*-  Mode:Common-Lisp; Base: 10; Package:cl-tests  -*-

(in-package :cl-tests)

; 5/16/90 mwp Abstracted all test ssystem support fns here from "00 A Testsystem"
; This file includes the guts of the testsystem: everything except 
;      - top level forms (to be called by user or test suite writer - these are in
;                         "00 A Testsystem"), and
;      - I/O (which are in TS I/O Fns.Lisp)

; the main test evaluating/error handling function
(defun errset (form &optional (flag nil))
  (declare (ignore flag))
;  (setf *current-error* nil)
  (if (or (equal *where-in-test?* :setup)
          (equal *where-in-test?* :unsetup))
    (setf form (seek-and-declare form nil)))  ;catches undeclared vars in setup and unsetup forms, makes 'em defparams
  (let ((val (ccl::handler-case (ccl::%eval-check-regs form) (error (c) c))))
     (if (typep val 'error)
       (progn
         (setf *current-error* val) nil)
       (list val))))



; keeps list of failed-tests
(defun record-error (name)
  (setf *failed-tests* (cons (cons name *testing-chapter*) *failed-tests*)))

; run-setup evals the setup form, returns non-nil if errset returns error, else nil
(defun run-setup (item setup name contrib$ failform error$ unsetup doc$)
  (setf *where-in-test?* :setup)
  (if (null (errset setup))
    (progn
      (report-error item setup name contrib$ failform error$ unsetup :setup doc$)
      :setup-error)
    nil))

; run-test returns non-nil if error, nil else
;      item is the test's first name (most general)
;      testform is, eg (equal (+ 1 2) 3)
;      name is the name of the specific test
;      type is :noeval (the default), :eval (if the 3rd element in testform should be evaled),
;              :error, :is-error 
;      failform is usually nil

(defun run-test (item testform name type contrib$ failform error$ unsetup doc$ tested)
  "Returns NIL if the test runs ok, else returns a keyword indicating where the test failed."
  (setf *where-in-test?* :run-test)
  (if *show-all-tests* (format t "~%~a - ~s - ~s" name *tests-counter* (print-time (get-universal-time))))
  (setf *tests-counter* (1+ *tests-counter*))
  (mapcar #'(lambda (x)(if (not (member x *tested-fns*))
                         (setf *tested-fns* (cons x *tested-fns*)))) tested)

  (let ((evalform :evalform-not-set-yet)
	(compareform :comp-form-not-set-yet)
	(result :no-result-yet)
        (predicate (first testform)))

    (cond ((eq type :error) 
           (cond ((setq evalform (errset testform))
                  (report-no-error name contrib$ failform error$ unsetup (car evalform) doc$)		 
                  :missing-error)))

	  ((eq type :is-error) 
	   (cond((setq evalform (errset testform))
		 (report-no-is-error name contrib$ failform error$ unsetup (car evalform) doc$)		 
		 :missing-is-error)))

          ; tests for "error in evalform" errors where no error expected
          ;            evalform is set to (3) - the evaluated (second testform) as a list
	  ((null (setq evalform (errset (second testform))))
	   (report-error item (second testform) name contrib$ failform error$ unsetup :evalform doc$)	   
	   :evalform)

          ;tests to make sure expected result is a legal form if it should be evaled (not type :noeval)
          ;            compareform set to (3)
	  ((null (setq compareform (cond ((eq type :noeval)
                                          (list (third testform)))
                                         (t (errset (third testform))))))
           (report-error item (third testform) name contrib$ failform error$ unsetup :compareform doc$)	   
	   :compareform)

          ; tests whether predicate is legal with evalform & compareform
          ;            result is set to (t) - the result of (errset '(apply #'equal '(3 3)))
	  ((null (setq result (errset `(apply (function ,predicate)
                                            '(,(car evalform) ,(car compareform))))))
    	   (report-error item testform name contrib$ failform error$ unsetup :testform :doc$)
	   :testform)

          ; reports test failure (but no error) if result not (t)
	  ((neq t (car result))	
   	   (report-test-fail item name (car result) testform doc$ error$ contrib$ (errset (car evalform)))
	   :test-failure))))

; run-unsetup evals unsetup form, returns non-nil if errset returns error, else nil
(defun run-unsetup (item unsetup name contrib$ failform error$ doc$)
  (setf *where-in-test?* :unsetup)
  (if (null (errset unsetup))
    (progn
      (report-error item unsetup name contrib$ failform error$ nil :unsetup doc$)
     :unsetup-error)
    nil))
	
(defun run-seq-setup (item setup name contrib$ doc$)
  (setf *where-in-test?* :setup)
  (if (null (errset setup))
    (progn
      (report-error item setup name contrib$ nil nil nil :seq-setup doc$)
      :seq-setup-error)
       nil))


(defun run-seq-unsetup (item unsetup name contrib$ doc$)
  (setf *where-in-test?* :unsetup)
  (if (null (errset unsetup))
    (progn
      (report-error item unsetup name contrib$ nil nil nil :seq-unsetup doc$)
	:seq-unsetup-error)
       :seq-completed))

(defun run-seq-test (item tests name contrib$ unsetup)
  (declare (ignore unsetup contrib$ name))
  (catch 'error-exit
    (dolist (a-test tests)
      (eval `(runtest (,item ,@(car a-test)) ,(second a-test) :from-seq t ,@(cddr a-test))))

;      (let ((value (eval `(runtest (,item ,@(car a-test)) ,(second a-test) :from-seq t ,@(cddr a-test)))))
;	(if (neq value :test-complete)(throw 'error-exit value))
    :seq-tests-completed))

(defun run-test-seq (item name setup unsetup test-seq contrib$ doc$)
  (let (result)
    (setq *sequence-p* name)
    (setq result (or (run-seq-setup item setup name contrib$ doc$)
		     (neq :seq-tests-completed (run-seq-test item test-seq name contrib$ unsetup))
		     (run-seq-unsetup item unsetup name contrib$ doc$)))
    (setq *sequence-p* nil)
    (cond ((neq result :seq-completed)
	   (format *test-error-stream*
		   "This last error early-terminates test-sequence ~s.~%"
		   name)
	   (if (neq result :seq-unsetup-error) (run-seq-unsetup item unsetup name contrib$ doc$))))
    result))



;-----------------------------------------------------------------------------
;;; Some generally useful functions.

; tests for non-nil instead of (equal t)
(defun runtest-eql (x y)
  (if (eq y <undefined>) t
      (or (and (eq y <true>) x t)
	  (eql x y))))

(DEFMACRO FUNCTION-TEST (X)
  "Test if X is a function."
  `(AND (FBOUNDP ,X) (NOT (MACRO-FUNCTION ,X)) (NOT (SPECIAL-FORM-P ,X))))

(defmacro makunboundlist (&rest list)
  (dolist (atom list)
    (makunbound atom)))

(defmacro fmakunboundlist (&rest list)
  (dolist (atom list)
    (fmakunbound atom)))

(defmacro unbindlist (functions symbols)
  `(progn (makunboundlist ,@ symbols)
	  (fmakunboundlist ,@ functions)))

(defun member-p (item list)
  (and (member item list) t))

;(defun neq (x y) (null (eq x y)))

#+ignore ;; this is completely wrong and breaks several tests -- Klmn 5/30/94
; walks through setup forms and unsetup forms, looking for setfs and setqs, changing
; (setf <var> <val> <var> <val> ...) to (progn (defparameter <var> <val>)(defparameter <var> <val>) ...)
; to avoid undeclared-free-variable compiler deaths.
(defun seek-and-declare (list result)
    (if (null list) result
        (if (listp (car list)) (progn 
                                   (setf list (cons (seek-and-declare (car list) nil)(cdr list)))
                                   (seek-and-declare (cdr list)(append result (list (car list)))))
              (if (or (equal (car list) 'setf)
                      (equal (car list) 'setq)) (seek-and-declare nil
                                          (append result (sets-to-defparams list)))
            (seek-and-declare (cdr list)(append result (list (car list))))))))

#+ignore
(defun sets-to-defparams (form )
  (let (new-l)
    (if (and (listp form)
             (or (equal (car form) 'setq)
                 (equal (car form) 'setf)))
      (do ((x 1 (+ x 2)))((>= x (length form))
                          (cons 'progn (reverse new-l)))
                     (setf new-l (cons (list 
                                        (if (symbolp (nth x form))
                                          (if (boundp (nth x form))
                                            'setf
                                            'defparameter)
                                          'setf)
                                             (nth x form)
                                             (nth (1+ x) form)) 
                          new-l)))
      form)))


;;this just returns the form.
(defun seek-and-declare (form result)
  result
  form)




;-----------------------------------------------------------------------------
