
/*
  Ensure that SystemExceptionHandler defers to application handler if installed.
  
02/01/97 gb  GCC tweaks.
--- 4.0
06/10/96 gb  new file.
--- 3.9 release

*/

#include <MachineExceptions.h>
#include <CodeFragments.h>
#include <Gestalt.h>
#include <MixedMode.h>
#include <Errors.h>

#define gestaltSaneSystemExceptionHandler 'SSXH'

typedef pascal ExceptionHandler (*HandlerInstallProc) (ExceptionHandler);

OSErr 
clobber_tv (void *dest, void *src)
{
  unsigned *d = (unsigned *)dest, *s = (unsigned *)src;
  
  d[0] = s[0];
  d[1] = s[1];
  
  return noErr;
}


ExceptionHandler
find_user_handler()
{
  ExceptionHandler user_handler = InstallExceptionHandler(NULL);
   
  InstallExceptionHandler(user_handler);
  return user_handler;
}

ExceptionHandler
real_sys_handler = NULL;

Ptr
lookup_cfm_symbol(ConstStr63Param libname, ConstStr255Param symname)
{
  Ptr value = NULL;
  CFragConnectionID connID;
  Str255 error_message;
  CFragSymbolClass symclass;
  Ptr mainaddr;
  OSErr err;
  
  err = GetSharedLibrary(libname, kPowerPCCFragArch, kFindLib, &connID, &mainaddr, error_message);
  if (err == noErr) {
    err = FindSymbol(connID, symname, &value, &symclass);
  }
  if (err != noErr) {
    return NULL;
  }
  return value;
}

OSStatus
sane_exception_handler(ExceptionInformationPowerPC *xp)
{
  OSStatus status;
  ExceptionHandler user_handler = find_user_handler();
   
  if (user_handler != NULL) {
    status = (user_handler)(xp);
    if (status == noErr) {
      return status;
    }
  }
   
  if (real_sys_handler != NULL) {
    return (real_sys_handler)(xp);
  }
   
  return (-1);
}

ExceptionHandler
shielded_system_install(ExceptionHandler new_handler)
{
  ExceptionHandler previous = real_sys_handler;
  
  real_sys_handler = new_handler;
  return previous;
}

OSErr
SSXH_gestalt_selector(OSType selector, long *response)
{
  if (selector != gestaltSaneSystemExceptionHandler) {
    return gestaltUnknownErr;
  }
  *response = 1;
  return noErr;
}

RoutineDescriptor
SSXH_gestalt_descriptor = BUILD_ROUTINE_DESCRIPTOR(uppSelectorFunctionProcInfo, SSXH_gestalt_selector);

SelectorFunctionUPP SSXH_Upp = &SSXH_gestalt_descriptor;

/*
  1) Find 'InstallSystemExceptionHandler'.
  2) Call it to install our 'sane' system exception handler, noting
  the old handler it returns.
  3) Bash InstallSystemExceptionHandler's transfer vector, so that
  it calls our version.
  4) Register a Gestalt selector, so that we can tell that this has
  already happened.
*/

#ifdef __MWERKS__
/*
  The MW linker uses this (magic name & all) to build the
  accelerated resource's header.
*/

ProcInfoType __procinfo = kCStackBased;
#endif

#ifdef __GNUC__
asm (".globl SSXHmain[DS]");
#endif

void
#ifdef __MWERKS__
main()
#else
SSXHmain()
#endif
{
  SelectorFunctionUPP oldGestaltFunction;
  HandlerInstallProc sys_install = NULL;
  
  sys_install = (HandlerInstallProc) 
    lookup_cfm_symbol("\pPrivateInterfaceLib", "\pInstallSystemExceptionHandler");
  if (sys_install != NULL) {
    real_sys_handler = ((sys_install)(sane_exception_handler));
    clobber_tv(sys_install, shielded_system_install);
    if (NewGestalt(gestaltSaneSystemExceptionHandler, SSXH_Upp) == gestaltDupSelectorErr) {
      ReplaceGestalt(gestaltSaneSystemExceptionHandler, SSXH_Upp, &oldGestaltFunction);
    }
  }
}
