;;	Change History (most recent first):s
;;  3 12/12/95 akh  gary's fix to tcallnfnslide
;;  (do not edit before this line!!)
;;; Subprims for various flavors of function call, tail-recursion, etc.

;;; 06/19/96 bill Add some comments about stack overflow assumptions.
;;; ------------- MCL-PPC 3.9
;;; 04/02/96 bill lexpr_entry initializes lisp_frame.savefn instead of lisp_frame.savelr twice
;;; 03/20/96 gb  .SPreset does (throw :toplevel $xstkover)
;;; 02/05/96 bill Gary's fix to lexpr_entry
;;; 01/29/96 gb  less hysteria in lexpr_entry
;;; 01/18/96 gb  add restorecontext, lexpr_entry
; 12/13/95  gb  udfcall->reset; callbuiltin, context stuff.
; 11/29/95  gb use jump_fname, jump_nfn macros
	include	'lisp.s'

	export funcall
	export jmpsym
	import ksignalerr



;;; Note: all the tXXXslide functions MUST use a variable other than VSP while pushing to the slid location.
;;; This causes do_vsp_overflow in "lisp-exceptions.c" to allocate a new segment for the overflow, ensuring
;;; that it won't overwrite the stuff being slid.
;;; (They have to do so anyway, so that the VSP protects everything during the slide, but someone may
;;; need to remember the do_vsp_overflow assumption sometime).

;;; Tail-recursively call the (known symbol) in fname.
	;; In the general case, we don't know if any args were
	;; vpushed or not.  If so, we have to "slide" them down
	;; to the base of the frame.  If not, we can just restore
	;; vsp, lr, fn from the saved lisp frame on the control stack.
tcallsymgen:	spentry tcallsymgen
	cmpwi cr0,nargs,nargregs<<fixnumshift
	ble cr0,tcallsymvsp
	;; b tcallsymslide

	;; Some args were vpushed.  Slide them down to the base of
	;; the current frame, then do funcall.
tcallsymslide:	spentry tcallsymslide
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz imm0,lisp_frame.savevsp(sp)
	la sp,lisp_frame.size(sp)
	mtlr loc_pc
	;; can use nfn (= temp2) as a temporary
	subi imm1,nargs,nargregs<<fixnumshift
	add imm1,imm1,vsp
@loop:	lwzu temp2,-4(imm1)
	cmpw cr0,imm1,vsp
	push temp2,imm0
	bne cr0,@loop
	mr vsp,imm0
	b jmpsym

	;; No args were vpushed; recover saved context & call symbol
tcallsymvsp:	spentry tcallsymvsp
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz vsp,lisp_frame.savevsp(sp)
	la sp,lisp_frame.size(sp)
	mtlr loc_pc
	;; b jmpsym

jmpsym:	spentry jmpsym
	jump_fname

;;; Tail-recursively call the function in nfn.
	;; Pretty much the same as the tcallsym* cases above.
tcallnfngen:	spentry tcallnfngen
	cmpwi cr0,nargs,nargregs<<fixnumshift
	ble cr0,tcallnfnvsp
	;; b tcallnfnslide

	;; Some args were vpushed.  Slide them down to the base of
	;; the current frame, then do funcall.
tcallnfnslide:	spentry tcallnfnslide
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz imm0,lisp_frame.savevsp(sp)
	la sp,lisp_frame.size(sp)
	mtlr loc_pc
	;; Since we have a known function, can use fname as a temporary.
	subi imm1,nargs,nargregs<<fixnumshift
	add imm1,imm1,vsp
@loop:	lwzu fname,-4(imm1)
	cmpw cr0,imm1,vsp
	push fname,imm0
	bne cr0,@loop
	mr vsp,imm0
	b jmpnfn

tcallnfnvsp:	spentry tcallnfnvsp
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz vsp,lisp_frame.savevsp(sp)
	la sp,lisp_frame.size(sp)
	mtlr loc_pc
	;; b jmpnfn

jmpnfn:	spentry jmpnfn
	jump_nfn


;;; Tail-recursively funcall temp0.
	;; Pretty much the same as the tcallsym* cases above.
tfuncallgen:	spentry tfuncallgen
	cmpwi cr0,nargs,nargregs<<fixnumshift
	ble cr0,tfuncallvsp
	;; b tfuncallslide

	;; Some args were vpushed.  Slide them down to the base of
	;; the current frame, then do funcall.
tfuncallslide:	spentry tfuncallslide
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz imm0,lisp_frame.savevsp(sp)
	la sp,lisp_frame.size(sp)
	mtlr loc_pc
	;; can use nfn (= temp2) as a temporary
	subi imm1,nargs,nargregs<<fixnumshift
	add imm1,imm1,vsp
@loop:	lwzu temp2,-4(imm1)
	cmpw cr0,imm1,vsp
	push temp2,imm0
	bne cr0,@loop
	mr vsp,imm0
	b funcall

	;; No args were vpushed; recover saved context & do funcall
tfuncallvsp:	spentry tfuncallvsp
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz vsp,lisp_frame.savevsp(sp)
	mtlr loc_pc
	la sp,lisp_frame.size(sp)
	;; b funcall

	;; Call temp0 if it's either a symbol or function
funcall: spentry funcall
	extract_lisptag imm0,temp0
	cmpwi imm0,tag_misc
	mr fname,temp0
	bne @error
	extract_subtag imm0,temp0
	cmpwi imm0,subtag_symbol
	cmpwi cr1,imm0,subtag_function
	bne cr0,@notsym
	jump_fname
@notsym:
	mr nfn,temp0
	bne cr1,@error
	jump_nfn
@error:	mr arg_z,temp0
	li arg_y,XNOTFUN	; not at all fun
	set_nargs 2
	b ksignalerr
	
reset:	spentry reset
	import .SPthrow
	nop
;;; rnil MUST be lisp_nil here.  When debugging, you
;;; should check that the low nibble of rnil (aka rtoc,
;;; aka r2) = #x5 or #xd.  You may be able to find
;;; the "real" value of lisp_nil from the pmcl-kernel
;;; debugger info file (pmcl-kernel.xcoff/pmcl-kernel.xsym);
;;; use some command named "show globals" or somesuch.
	ref_nrs_value temp0,toplcatch
	li temp1,XSTKOVER
	vpush temp0
	vpush temp1
	set_nargs 1
	b .SPthrow
	
	

;;; the value of the nilreg-relative symbol %builtin-functions% should be
;;; a vector of symbols.  Funcall the symbol indexed by imm0 (boxed) and
;;; return a single value.

callbuiltin0: spentry callbuiltin0
	set_nargs 0
	b callbuiltin

callbuiltin1: spentry callbuiltin1
	set_nargs 1
	b callbuiltin

callbuiltin2: spentry callbuiltin2
	set_nargs 2
	b callbuiltin

callbuiltin3: spentry callbuiltin3
	set_nargs 3
callbuiltin: spentry callbuiltin
	ref_nrs_value fname,builtin_functions
	la imm0,misc_data_offset(imm0)
	lwzx fname,fname,imm0
	jump_fname

popj:	spentry popj
	export .popj
.popj:	
	lwz loc_pc,lisp_frame.savelr(sp)
	mtlr loc_pc
	lwz vsp,lisp_frame.savevsp(sp)
	lwz fn,lisp_frame.savefn(sp)
	la sp,lisp_frame.size(sp)
	blr

restorefullcontext: spentry restorefullcontext
	mflr loc_pc
	mtctr loc_pc
	lwz loc_pc,lisp_frame.savelr(sp)
	mtlr loc_pc
	lwz vsp,lisp_frame.savevsp(sp)
	lwz fn,lisp_frame.savefn(sp)
	la sp,lisp_frame.size(sp)
	bctr	

;;; Like .SPrestorefullcontext, only the saved return address
;;; winds up in loc-pc instead of getting thrashed around ...
restorecontext:	spentry restorecontext
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz vsp,lisp_frame.savevsp(sp)
	lwz fn,lisp_frame.savefn(sp)
	la sp,lisp_frame.size(sp)
	blr
	
	import cscheck
savecontextvsp: spentry savecontextvsp
	li imm0,0
	; b savecontext0

savecontext0: spentry savecontext0
	stwu sp,-lisp_frame.size(sp)
	add imm0,vsp,imm0
	stw fn,lisp_frame.savefn(sp)
	stw loc_pc,lisp_frame.savelr(sp)
	stw imm0,lisp_frame.savevsp(sp)
	mr fn,nfn
	b cscheck

;;; Nargs is valid; all arg regs, lexpr-count pushed by caller.
;;; imm0 = vsp to restore.
;;; Return all values returned by caller to its caller, hiding
;;; the variable-length arglist.
;;; If we can detect that the caller's caller didn't expect
;;; multiple values, then things are even simpler.
lexpr_entry: spentry lexpr_entry
	ref_global imm1,ret1val_addr
	cmpw cr0,imm1,loc_pc
	stwu sp,-lisp_frame.size(sp)
	stw fn,lisp_frame.savefn(sp)
	stw loc_pc,lisp_frame.savelr(sp)
	stw imm0,lisp_frame.savevsp(sp)
	bne cr0,@single_value
	ref_global imm0,lexpr_return
	stwu sp,-lisp_frame.size(sp)
	stw rzero,lisp_frame.savefn(sp)
	stw imm0,lisp_frame.savelr(sp)
	stw vsp,lisp_frame.savevsp(sp)
	mr loc_pc,imm1
	li fn,0
	b cscheck

;;; The single-value case just needs to return to something that'll pop
;;; the variable-length frame off of the vstack.
@single_value:
	ref_global loc_pc,lexpr_return1v
	li fn,0
	b cscheck

	
	
	end
